/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *  TOKENS.CPP -- token class                          V1.0  *
 *                                                           *
 *  Language:      Borland C++                               *
 *  System:        Gateway 386/25, MS-DOS 5.0                *
 *  Application:   CSE 687                                   *
 *  Author:        Jim Fawcett, CST 2-127, (315)443-4454     *
 *                                                           *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#include <stdlib.h>        // ANSI:     exit
#include <ctype.h>         // ANSI:     isspace
#include "tokens.hpp"      // FAWCETT:  prototypes for tokens

#define TOKSIZE 1024

//----< token::token(void) >---------------------------------//

token::token(void) {

  delims = " \n\t\f\r\b";   // default delimit on whitespace
  top_position = position = prev_position = 0;
  lines = 0;
}

/*----< token::token(char *filename) >-----------------------*/

token::token(char *filename) {

  ifstrm.open(filename,ios::in);
  if (!ifstrm.good()) {
    cerr << "can't open " << filename << "\n";
    exit(1);
  }
  delims = " \n\t\f\r\b";   // default delimit on whitespace
  top_position  = position = ifstrm.tellg();
  prev_position = 0;
  lines = 0;
}

/*----< token::~token(void) >--------------------------------*/

token::~token(void) {

  ifstrm.close();
}

//----< int get_tok(token &tok) >----------------------------//

int get_tok(token &tok) {

  char *whitespace     = " \n\t\f\b\v";
  char *quote_or_paren = "(\"\'";
  int ch, i;
  char work[TOKSIZE];

// check end of file
  if (tok.ifstrm.eof())
    return 0;
  else {
  // eat whitespace
    while( is_in(ch=tok.ifstrm.get(), whitespace)) {
      if (ch == '\n') tok.lines++;
      if (ch == EOF)
	return 0;
    }
    tok.ifstrm.putback(ch);  // first non-whitespace char
  // collect token
    i=0;
  // if first non-whitespace char is delim it is a token
    if (is_in(ch=tok.ifstrm.get(), tok.delims)) {
      work[0] = ch;
      work[1] = '\0';
      return 1;
    }
    else      // put it back and start building token
      tok.ifstrm.putback(ch);
  // collect token chars
    while(!is_in(ch=tok.ifstrm.get(), tok.delims)) {
      if (ch == '\n') tok.lines++;
      work[i++] = ch;
      if (ch == EOF) break;
      if((work[i-2]=='/')&&(work[i-1]=='*'))   // C comment
	break;
      if((work[i-2]=='/')&&(work[i-1]=='/'))   // C++ comment
	break;
//    if ((work[i-1]=='(')||(work[i-1]=='\"')) // parens or quotes
      if (is_in(work[i-1],quote_or_paren))
	if (i==1)
	  break;
	else {
	  tok.ifstrm.putback(ch);
	  i--;
	  break;
	}
    }
    if (ch == '\n') tok.lines++;
    work[i] = '\0';
  }
  tok.str = work;
  if (tok.str == "/*")
    collect_comments(tok, C_code);
  else if (tok.str == "//")
    collect_comments(tok, Cpp_code);
  else if (tok.str[i-1] == '(')
    collect_parens(tok);
  else if (tok.str[i-1] == '\"')
    collect_quotes(tok,'\"');
  else if (tok.str[i-1] == '\'')
    collect_quotes(tok,'\'');
  return 1;
}

//----< int next(token &tok) >-------------------------------//

int next(token &tok) {

  tok.prev_tok = tok.str;
  return get_tok(tok);
}

//----< int prev(token &tok) >------------------------------//

int prev(token &tok) {

  if (length(tok.prev_tok) <= 0)
    return 0;
  tok.str = tok.prev_tok;
  return 1;
}

/*----< int done(token &tok) >------------------------------*/

int done(token &tok) {

  if (tok.ifstrm.eof())  // ios::eof() returns nonzero on eof
    return 1;
  return 0;
}

//----< void rewind(token &tok) >---------------------------//

void rewind(token &tok) {

  tok.ifstrm.seekg(tok.top_position);
  tok.ifstrm.clear();
}

/*----< string collect_parens(token &tok) >-----------------*/

int collect_parens(token &tok) {

  char work[TOKSIZE+1];
  int ch, quote_char, in_quotes=0, j=1;
  for (int i=0; i<TOKSIZE; i++) {
    ch = tok.ifstrm.get();
    if (ch == '\n') tok.lines++;
    if ((ch == '\"')||(ch == '\'')) {  // toggle in_quotes
      if (!in_quotes) {
	quote_char = ch;
	in_quotes = 1;
      }
      else if (ch == quote_char)
	in_quotes = 0;
    }
    if (ch == EOF)
      break;
    else
      work[i] = ch;
    if ((ch == '(')&&(!in_quotes)) j++;
    else if ((ch == ')')&&(!in_quotes)) j--;
    if (j==0) {
      work[++i] = '\0';
      break;
    }
  }
  work[TOKSIZE] = '\0';
  tok.str += work;
  return 1;
}

//----< string collect_comments(token &tok, code type) >-----//

//      Collect comment characters and load into tok.
//      Assumes "/*" already in tok.

int collect_comments(token &tok, code type) {

  char work[TOKSIZE+1], ch;
  if (type == Cpp_code) {
    for (int i=0; i<TOKSIZE; i++) {
      if ((ch = work[i] = tok.ifstrm.get()) == '\n') {
	tok.lines++;
	break;
      }
      if (ch == EOF) break;
    }
    work[i] = '\0';
    tok.str += work;
    return 1;
  }
  while (tok.ifstrm.good()) {
    for (int i=0; i<TOKSIZE; i++) {
      work[i] = tok.ifstrm.get();
      if (work[i] == '\n') tok.lines++;
      if ((work[i-1] == '*')&&(work[i] == '/')) {
	work[i+1] = '\0';
	tok.str += work;
	return 1;
      }
      if (work[i] == EOF) {
	work[i] = '\0';
	tok.str += work;
	return 0;
      }
    }
    work[TOKSIZE] = '\0';
    tok.str += work;
  }
  return 1;
}

//----< string collect_quotes(token &tok, char qch) >--------//

int collect_quotes(token &tok, char qch) {

  char work[TOKSIZE+1];
  int ch;
  for (int i=0; i<TOKSIZE; i++) {
    ch = tok.ifstrm.get();
    if (ch == '\n') tok.lines++;
    if (ch == EOF)
      break;
    else
      work[i] = ch;
    if (ch == qch) {
      work[++i] = '\0';
      break;
    }
  }
  work[TOKSIZE] = '\0';
  tok.str += work;
  return 1;
}
